import aiohttp
import asyncio
import logging
from typing import Optional, Callable, Awaitable, List
import urllib
from urllib.parse import urlencode
from multidict import MultiDict


from . import exceptions, common
from datetime import datetime
from decimal import Decimal

__all__ = ('run')


logger = logging.getLogger(__name__)


MarketDataCallback = Callable[[dict], Awaitable[None]]
OrderBookCallback = Callable[[int, str, dict, dict], Awaitable[None]]
TradesCallback = Callable[[datetime, int, str, Decimal, Decimal], Awaitable[None]]


async def reader_loop(
        ws: aiohttp.ClientWebSocketResponse,
        market_data_callback: MarketDataCallback,
        order_book_callback: OrderBookCallback,
        trades_callback: TradesCallback) -> None:
    logger.info(f'broadcast connection established')
    while True:
        msg = await common.receive_msg(ws)

        try:
            message_type: common.ServerMessageType = common.ServerMessageType[msg['response_type']]
            if message_type != common.ServerMessageType.BROADCAST:
                raise exceptions.UnsupportedMessageType()
            payload = msg['data']
            if market_data_callback is not None:
                await market_data_callback(payload)
            if payload['@type'] == 'OrderBookAgg':
                if order_book_callback is not None:
                    asyncio.ensure_future(order_book_callback(
                        payload['current_order_id'],
                        payload['trade_pair'],
                        payload.get('buy_levels', dict),
                        payload.get('sell_levels', dict)
                    ))
            elif payload['@type'] == 'AnonymousTrade':
                if trades_callback is not None:
                    asyncio.ensure_future(trades_callback(
                        datetime.utcfromtimestamp(payload['time'][0]),
                        payload['current_order_id'],
                        payload['trade_pair'],
                        Decimal(payload['amount']),
                        Decimal(payload['price'])
                    ))
            else:
                raise exceptions.UnsupportedMessageType()
        except (KeyError, ValueError, exceptions.UnsupportedMessageType):
            logger.exception('failed to decode data')
            raise exceptions.CryptologyError('failed to decode data')


async def run(*, ws_addr: str, market_data_callback: MarketDataCallback = None,
              order_book_callback: OrderBookCallback = None,
              trades_callback: TradesCallback = None,
              trade_pairs: List[str] = None,
              loop: Optional[asyncio.AbstractEventLoop] = None) -> None:
    url = ws_addr
    if trade_pairs:
        params = MultiDict()
        for trade_pair in trade_pairs:
            params.add('trade_pair', trade_pair)
        url = '{}?{}'.format(url, urlencode(params))
    async with aiohttp.ClientSession(loop=loop) as session:
        async with session.ws_connect(url, receive_timeout=20, heartbeat=3) as ws:
            await reader_loop(ws, market_data_callback, order_book_callback, trades_callback)



import asyncio
from typing import Awaitable, Iterable, Optional

__all__ = ('run_parallel',)


async def run_parallel(coros: Iterable[Awaitable[None]],
                       *,
                       raise_canceled: bool = False,
                       loop: Optional[asyncio.AbstractEventLoop] = None) -> None:
    """
    run coros in parallel, cancel all on first exit
    raises first non-canceled exception
    may raise `asyncio.CanceledError` when `raise_canceled` is set
    """
    tasks = list(asyncio.Task(x, loop=loop) for x in coros)

    if not tasks:
        return

    def cancel_others(target: asyncio.Task) -> None:
        for other in tasks:
            other.remove_done_callback(cancel_others)
        for other in tasks:
            if other is not target:
                other.cancel()

    for task in tasks:
        task.add_done_callback(cancel_others)

    result = await asyncio.gather(*tasks, return_exceptions=True, loop=loop)

    exception = None
    for err in filter(None, result):
        skip_canceled = exception is not None or not raise_canceled
        is_canceled = isinstance(err, asyncio.CancelledError)
        if is_canceled and skip_canceled:
            continue
        exception = err
        if not is_canceled:
            break

    if exception is not None:
        raise exception



import aiohttp
from cryptology import common


class CryptologyError(Exception):
    pass


class CryptologyProtocolError(CryptologyError):
    pass


class InvalidServerAddress(CryptologyProtocolError):
    pass


class IncompatibleVersion(CryptologyProtocolError):
    pass


class InvalidKey(CryptologyProtocolError):
    pass


class InvalidSequence(CryptologyProtocolError):
    pass


class UnsupportedMessage(CryptologyProtocolError):
    msg: aiohttp.WSMessage

    def __init__(self, msg: aiohttp.WSMessage) -> None:
        super(CryptologyProtocolError, self).__init__(f'unsupported message {msg!r}')
        self.msg = msg


class UnsupportedMessageType(CryptologyProtocolError):
    pass


class CryptologyConnectionError(CryptologyError):
    pass


class Disconnected(CryptologyConnectionError):
    def __init__(self, code: int, message: str) -> None:
        if message:
            super().__init__(f'disconnected with code {code} and message "{message}"')
        else:
            super().__init__(f'disconnected with code {code}')


class ConcurrentConnection(CryptologyConnectionError):
    pass


class ServerRestart(CryptologyConnectionError):
    pass


class RateLimit(CryptologyError):
    pass


class InvalidPayloadError(CryptologyProtocolError):
    pass


class PermissionDeniedError(CryptologyError):
    pass


class DuplicateClientOrderIdError(CryptologyProtocolError):
    pass


def handle_close_message(msg: aiohttp.WSMessage) -> None:
    if msg.type in common.CLOSE_MESSAGES:
        if msg.type == aiohttp.WSMsgType.CLOSE:
            if msg.data == 4000:
                raise ConcurrentConnection()
            elif msg.data == 4001:
                raise InvalidSequence()
            elif msg.data == 4009:
                raise RateLimit()
            elif msg.data == 1012:
                raise ServerRestart()
            elif msg.data == 4100:
                raise InvalidKey()
            elif msg.data == 4102:
                raise PermissionDeniedError(msg.extra)
            elif msg.data == 4103:
                raise IncompatibleVersion(msg.extra)
            elif msg.data == 4010:
                raise InvalidPayloadError(msg.extra)
            elif msg.data == 4013:
                raise PermissionDeniedError(msg.extra)
            elif msg.data == 4014:
                raise DuplicateClientOrderIdError(msg.extra)
        raise Disconnected(msg.data, msg.extra)



import sys

from setuptools import setup

if sys.version_info < (3, 6):
    raise ImportError('cryptology-ws-client-python only supports python3.6 and newer')

setup(
    name='cryptology-ws-client',
    version='1.1.1',
    description='Cryptology webscoket client',
    author='Cryptology',
    author_email='s.prikazchikov@cryptology.com',
    packages=['cryptology'],
    python_requires='>= 3.6',
    install_requires=[
        'aiodns',
        'aiohttp >= 2.3.6',
        'cchardet',
    ],
    extras_require={
        'devel': ['pytz',
                  'pytest-aiohttp'
                  ]
    },
    url='https://github.com/CryptologyExchange/cryptology-ws-client-python',
)



import json
import logging
from datetime import timedelta
from enum import Enum, unique
from typing import Any, Optional

import aiohttp

from . import exceptions


logger = logging.getLogger(__name__)


HEARTBEAT_INTERVAL = timedelta(seconds=2)

CLOSE_MESSAGES = (aiohttp.WSMsgType.CLOSE, aiohttp.WSMsgType.CLOSING,
                  aiohttp.WSMsgType.CLOSED, aiohttp.WSMsgType.ERROR,)


class ByValue(Enum):
    @classmethod
    def by_value(cls, value: int) -> Any:
        for val in cls:
            if val.value == value:
                return val
        raise IndexError(value)


@unique
class ClientMessageType(ByValue):
    INBOX_MESSAGE = 1
    RPC_REQUEST = 2


@unique
class ServerMessageType(ByValue):
    MESSAGE = 1
    ERROR = 2
    BROADCAST = 3
    THROTTLING = 4


class ServerErrorType(ByValue):
    UNKNOWN_ERROR = -1
    DUPLICATE_CLIENT_ORDER_ID = 1
    INVALID_PAYLOAD = 2
    PERMISSION_DENIED = 3


async def receive_msg(ws: aiohttp.ClientWebSocketResponse, *, timeout: Optional[float] = None) -> dict:
    msg = await ws.receive(timeout=timeout)
    if msg.type in CLOSE_MESSAGES:
        logger.info('close msg received (type %s): %s', msg.type.name, msg.data)
        exceptions.handle_close_message(msg)
        raise exceptions.UnsupportedMessage(msg)

    return json.loads(msg.data)



import aiohttp
import asyncio
import functools
import inspect
import logging

from datetime import datetime
from typing import Any, AsyncIterator, Awaitable, Callable, ClassVar, Optional, Tuple, Type, cast, Dict, List

from . import common, exceptions, parallel


__all__ = ('ClientReadCallback', 'ClientWriter', 'ClientWriterStub', 'run_client',)

logger = logging.getLogger(__name__)


CLIENTWEBSOCKETRESPONSE_INIT_ARGS = list(
    inspect.signature(aiohttp.ClientWebSocketResponse.__init__).parameters.keys())[1:]


class ClientWriterStub:
    async def send_message(self, *, payload: dict) -> None:
        pass


ClientReadCallback = Callable[[ClientWriterStub, datetime, int, dict], Awaitable[None]]
ClientWriter = Callable[[ClientWriterStub, List[str], Optional[Dict]], Awaitable[None]]
ClientThrottlingCallback = Callable[[int, int], Awaitable[bool]]


class BaseProtocolClient(aiohttp.ClientWebSocketResponse):
    VERSION: ClassVar[int] = 7

    access_key: ClassVar[str]
    secret_key: ClassVar[str]

    sequence_id: int

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        kw = {}
        kw.update(dict(zip(CLIENTWEBSOCKETRESPONSE_INIT_ARGS, args)))
        kw.update(kwargs)
        super(BaseProtocolClient, self).__init__(**kw)
        self.send_fut = None
        self.throttle = 0

    async def authenticate(self, last_seen_message_id: int, get_balances: bool = False,
                           get_order_books: bool = False) -> Tuple[int, int, Dict, List[str]]:
        state_request_data = {}
        if get_balances:
            state_request_data['get_balances'] = get_balances
        if get_order_books:
            state_request_data['get_order_books'] = get_order_books

        await self.send_json({'access_key': self.access_key,
                              'secret_key': self.secret_key,
                              'last_seen_message_id': last_seen_message_id,
                              'version': self.VERSION,
                              **state_request_data})
        data = await common.receive_msg(self)
        try:
            if data['greeting'] != 'Welcome to Cryptology API Server':
                raise exceptions.InvalidServerAddress()
            last_seen_sequence = data['last_seen_sequence']
            server_version = data['version']
            state = data.get('state')
            pairs = data['trade_pairs']
        except (KeyError, TypeError):
            raise exceptions.InvalidServerAddress()
        self.sequence_id = last_seen_sequence
        return last_seen_sequence, server_version, state, pairs

    async def send_message(self, *, payload: dict) -> None:
        if self.closed:
            logger.warning('the socket is closed')
            raise exceptions.CryptologyConnectionError()

        self.sequence_id += 1
        sequence_id = self.sequence_id
        data = {'sequence_id': sequence_id,
                'data': payload}
        if self.send_fut:
            await self.send_fut
        if self.throttle:
            logger.warning('throttle for %f seconds', self.throttle)
            throttle, self.throttle = self.throttle, 0
            await asyncio.sleep(throttle)
        logger.debug('sending message with seq id %i: %s', sequence_id, payload)
        self.send_fut = asyncio.ensure_future(self.send_json(data))

    async def receive_iter(self, throttling_callback: ClientThrottlingCallback) -> AsyncIterator[Tuple[datetime, dict]]:
        while True:
            data = await common.receive_msg(self)

            message_type: common.ServerMessageType = common.ServerMessageType[data['response_type']]
            logger.debug('message %s received', message_type)
            if message_type is common.ServerMessageType.THROTTLING:
                level = data['overflow_level']
                sequence_id = data['sequence_id']
                if not throttling_callback or not await throttling_callback(level, sequence_id):
                    self.throttle = 0.001 * level
            elif message_type is common.ServerMessageType.MESSAGE:
                ts = data['timestamp']
                logger.debug('outbox message: %s', data['data'])
                yield datetime.utcfromtimestamp(ts), data['message_id'], data['data']
            else:
                logger.error('unsupported message type')
                raise exceptions.UnsupportedMessageType()

    def _pong_not_received(self):
        logger.warning('missed heartbeat from the server. connection will be closed')
        super()._pong_not_received()


@functools.lru_cache(typed=True)
def bind_response_class(access_key: str, secret_key: str) -> Type[BaseProtocolClient]:
    return cast(Type[BaseProtocolClient],
                type('BoundProtocolClient', (BaseProtocolClient,),
                     {'access_key': access_key, 'secret_key': secret_key}))


class CryptologyClientSession(aiohttp.ClientSession):
    def __init__(self, access_key: str, secret_key: str, *,
                 loop: Optional[asyncio.AbstractEventLoop] = None) -> None:
        super().__init__(ws_response_class=bind_response_class(access_key, secret_key), loop=loop, conn_timeout=10)


async def run_client(*, access_key: str, secret_key: str, ws_addr: str,
                     read_callback: ClientReadCallback, writer: ClientWriter,
                     throttling_callback: ClientThrottlingCallback = None,
                     last_seen_message_id: int = 0,
                     loop: Optional[asyncio.AbstractEventLoop] = None,
                     get_balances: bool = False,
                     get_order_books: bool = False,
                     error_callback: Any = None) -> None:
    if error_callback:
        logger.warning('error_callback is deprecated')
    async with CryptologyClientSession(access_key, secret_key, loop=loop) as session:
        async with session.ws_connect(ws_addr, autoclose=True, autoping=True, receive_timeout=10, heartbeat=4) as ws:
            logger.info('connected to the server %s', ws_addr)
            sequence_id, server_version, state, pairs = await ws.authenticate(last_seen_message_id,
                                                                              get_balances,
                                                                              get_order_books)
            logger.info('Authentication succeeded, server version %i, sequence id = %i', server_version, sequence_id)
            if server_version < 6:
                raise exceptions.IncompatibleVersion('Server version less than 6 is not supported')

            async def reader_loop() -> None:
                async for ts, message_id, msg in ws.receive_iter(throttling_callback):
                    logger.debug('%s new msg from server @%i: %s', ts, message_id, msg)
                    asyncio.ensure_future(read_callback(ws, ts, message_id, msg))

            await parallel.run_parallel((
                reader_loop(),
                writer(ws, pairs, state)
            ), loop=loop)
